/* $HeadURL$
* $LastChangedRevision$
* $LastChangedDate$
* $LastChangedBy$
*
* Description: DCF77 Sampling und Auswertung
* Code mit freundlicher Genehmigung von Georg-Johann Lay
* http://www.gjlay.de/software/c-code/dcf77.html
* 
*/

/*
Copyright (C) 20?? Georg-Johann Lay 
Extensions in 2012 by G. Herzig / DH8GHH

This program is free software; you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation; either version 3 of the License,
or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see <http://www.gnu.org/licenses/>
*/

#include "dcf77.h"
#include <avr/io.h>
#include "main.h"
#include "hardware.h"

#if defined (__GNUC__) && defined (__AVR__) && !defined (__FLASH)
#warning You need avr-gcc 4.7 or higher to use __flash!
#warning ...locating dcf_byteno[] in RAM
// Notlsung: Standard-C wird dcf_byteno[] im RAM speichern
#define __flash
#endif // avr-gcc

dcf77_t dcf77;

#define PARITY 0xff
#define IGNORE 0xfe
#define NEWVAL 0x10

// Adressdifferenz von .N zu .minute in time_t

#define X(N) ((uint8_t)(&dcf77.newtime.N - &dcf77.newtime.minute))

#define X0 X(minute)
#define X1 X(stunde)
#define X2 X(day)
#define X3 X(day_of_week)
#define X4 X(month)
#define X5 X(year)
#define X6 X(mesz)

/* Codiert die DCF-Bits 17..58 bzw. 21...58 (je nachdem, ob MESZ
   empfangen werden soll) und deren Position in time_t.
   Low-Nibble     : Offset relativ zu .minute
   Bit 4 (NEWVAL) : Neuer Wert fngt an, d.h. die Speicherposition
                    muss 1 Byte voranschreiten. */
static const __flash uint8_t dcf_byteno[] =
{
#ifdef TIME_MESZ
    // Zeitzone: MEZ/MESZ (Winter-/Sommerzeit): 2 Bits an Offset 6
    NEWVAL | X6, X6,
    // Ankndigung Schaltsekunde
    IGNORE,
    // Parity Zeitzone (immer 1)
    PARITY,
#endif // TIME_MESZ

    // Minute: 7 Bits an Offset 0
    NEWVAL | X0, X0, X0, X0, X0, X0, X0,
    // Parity Minute
    PARITY,
    // Stunde: 6 Bits an Offset 1
    NEWVAL | X1, X1, X1, X1, X1, X1,
    // Parity Stunde
    PARITY,
    // Tag: 6 Bits an Offset 2
    NEWVAL | X2, X2, X2, X2, X2, X2,
    // Wochentag: 3 Bits an Offset 3
    NEWVAL | X3, X3, X3,
    // Monat: 5 Bits an Offset 4
    NEWVAL | X4, X4, X4, X4, X4,
    // Jahr: 8 Bits an Offset 5
    NEWVAL | X5, X5, X5, X5, X5, X5, X5, X5,
    // Parity Datum
    PARITY
};

bool IsDCFSync(void)
{
	return false;
	//return (dcf77.error == 1);
}

static void dcf77_start_bit (dcf77_t*, uint8_t);
static void dcf77_store_bit (dcf77_t*, uint8_t);
static void dcf77_end_bit (dcf77_t*, uint8_t);

/*
   Dies ist der Einstiegspunkt fr das DCF-Modul. Die Funktion wird alle
   10ms (ein Tick) aufgerufen.

   poll = 0: Trger ist abgesenkt (erste 100ms oder 200ms einer Sekunde)
   poll = 1: Voller Trger */

void dcf77_tick (dcf77_t *dcf, uint8_t poll)
{
    // Diese Funktion wird zu jedem Tick -- also alle 10ms --
    // aufgerufen. Ticks um 1 weiterzhlen.

    uint8_t ticks = 1 + dcf->ticks;

    // berlauf der Ticks, z.B. bei schlechtem Signal oder nicht-
    // angeschlossenem Empfnger. Ticks bleiben dann bei 255 stehen.

    if (ticks == 0)
        ticks = UINT8_MAX;

    dcf->ticks = ticks;

    // poll normieren auf 0 oder !0
    poll = !!poll;

    // Flanke?
    if (poll != dcf->poll)
    {
        // Ja. poll-Wert zur Erkennung der nchsten Flanke merken.
        dcf->poll = poll;

        if (poll)
        {
            // Flanke 0->1: Bit-Ende 100ms oder 200ms nach Sekundenstart
            dcf77_end_bit (dcf, ticks);
        }
        else
        {
            // Flanke 1->0: Bit-Anfang, Sekundenstart
            dcf77_start_bit (dcf, ticks);
            dcf->ticks = 0;
        }
    }
}


/* Wird aufgerufen bei Flanke 1->0 des DCF-Signals.
   Start eines neuen Bits bzw. Start/Ende einer Sekunde.
   TICKS sind die Anzahl Ticks seit der vorherigen 1->0 Flanke
   und ist ca. 100 (also etwa 1s) oder nach dem ausgelassenen
   59. Bit ca. 200 (also etwa 2s). */

static void dcf77_start_bit (dcf77_t *dcf, uint8_t ticks)
{
    // Aktuelle Sekunde 0..59 um 1 weiterzhlen
    uint8_t second = dcf->newtime.sekunde;

    if (second < 60)
        second++;

    dcf->newtime.sekunde = second;

    if (ticks > 200-10 && ticks < 200+10)
    {
        // 59. Sekunde: Der Trger wurde fr ca. 200 Ticks (2s)
        // nicht abgesenkt. Nun beginnt eine neue Absenkung, d.h.
        // eine neue Minute beginnt.

        dcf->newtime.sekunde = 0;

        if (dcf->error == 0 && second == 59)
        {
            // Whrend einer ganzen Minute wurden korrekte
            // Bits empfangen --> aktuelle Zeit/Datum speichern

            dcf->time_changed = 1;

            // !!! FR .ptime SIEHE DEN KOMMENTAR IN dcf77.h     !!!

            if (dcf->ptime)
			{
				// UTC-Offset berechnen
				uint8_t offset = 23;		// Standard ist Winterzeit
				if (dcf->newtime.mesz == 1)	// If Sommerzeit, noch ne Stunde weniger
				{
					offset--;
				}
				dcf->newtime.stunde += offset;
				if (dcf->newtime.stunde > 23)
					dcf->newtime.stunde -= 24;

                *(dcf->ptime) = dcf->newtime;
				
			}
        }

        // Eine Minuten-Marke bewirkt ein Neubeginn der
        // DCF-Auswertung; daher das Fehler-Flag zurcksetzen.

        dcf->error = 0;
    }
    else
    {
        // Eine "normale" Sekunde ist vergangen. Deren Lnge
        // ist ca. 100 Ticks (1s). Andernfalls ist's ein Fehler.

        if (ticks < 100-10 || ticks > 100+10)
            dcf->error = 1;
    }
}


/* Wird aufgerufen bei Flanke 0->1 des DCF-Signals.
   Ende der Trgerabsenkung, d.h. Bitende. TICKS ist
   die Dauer der Trgerabsenkung, d.h. die Anzahl der
   10ms-Ticks seit der letzten 1->0 Flanke des DCF-Signals
   und ist fr gltige 0-Bits ca. 10 (100ms) und fr
   gltige 1-Bits ca. 20 (200ms). */

static void dcf77_end_bit (dcf77_t *dcf, uint8_t ticks)
{
    uint8_t bit = 0;

    if (ticks >= 10-3 && ticks < 10+3)
        // Ein 0-Bit dauert ca. 10 Ticks (100ms)
        bit = 0;
    else if (ticks >= 20-3 && ticks < 20+3)
        // Ein 1-Bit dauert ca. 20 Ticks (200ms)
        bit = 1;
    else
        // Alle anderen Bitlngen sind ungltig --> Fehler
        dcf->error = 1;

//DEBUG
	//time.stunde = ticks;
	//time_update = 1;
//ENDEBUG

    if (!dcf->error)
        dcf77_store_bit (dcf, bit);
}


/* Speichert BIT in der DCF-Struktur */

static void dcf77_store_bit (dcf77_t *dcf, uint8_t bit)
{
    // Nur Bits 17..58 bzw. 21..58 decodieren. Die aktuelle Sekunde,
    // vermindert um 17/21, dient als Index ins Feld dcf_byteno[],
    // wo Informationen dazu stehen, wie BIT zu
    // behandeln und wo es ggf. zu speichern ist.

    const uint8_t nbits = sizeof (dcf_byteno) / sizeof (dcf_byteno[0]);

    uint8_t i = dcf->newtime.sekunde - (59-nbits);

    if (i >= nbits)
        return;

    i = dcf_byteno[i];

#ifdef TIME_MESZ
    if (i == IGNORE)
        return;
#endif // TIME_MESZ

    dcf->parity ^= bit;

    if (i == PARITY)
    {
        // Partity-Test. Diese Funktion wird nur aufgerufen mit
        // .error = 0. Falls parity != 0 ist, liegt also ein
        // Parity-Fehler vor. Ansonsten bleibt .error = 0 unverndert.
        dcf->error = dcf->parity;

        // Parity fr nchstes Bit zurcksetzen.
        dcf->parity = 0;

        return;
    }

    // Low-Nibble -> Offset des Bytes (relativ zu .minute),
    // zu dem das Bit gehrt
    uint8_t offset = i & 0xf;
    uint8_t *d = &dcf->newtime.minute + offset;

    uint8_t val = *d;

    // Wertigkeit des Bits in BCD: 1, 2, 4, 8, 10, 20, 40, 80
    uint8_t bitval = dcf->bitval << 1;
    if (bitval == 0x10)
        bitval = 10;

    // Ein neuer Wert fngt an
    if (i & NEWVAL)
        val = 0, bitval = 1;

    dcf->bitval = bitval;

    // Falls DCF-Bit gesetzt ist: seine Wertigkeit zum
    // Zielbyte hinzuaddieren

    if (bit)
        val += bitval;

    *d = val;

    // Kleiner Konsistenztest: Testet Einer auf gltiges BCD-Format
    if (bitval == 8  &&  val >= 10)
        dcf->error = val;
}



// Krams aus der Main

// Merkt alle 10 Millisekunden den Portzustand des DCF-Ports
uint8_t volatile dcf77_bit = 0xff;


// Alle 10 Millisekunden aus einer ISR heraus aufrufen
// Macht alle 10ms einen Schnappschuss vom DCF-Port (hier an Port B1).
void job77_dcf_10ms (void)
{

    uint8_t bit = 1;

	uint8_t DoSound = 0;
	if ((Settings.SyncSound) && (!GetHupe()))
	{
		DoSound = 1;
	}

	if (DoSound)
	{
		PORTD &= ~(1<<PD5); //DEBUG
	}
    if (PIND & (1 << PD2))
	{
        bit = 0;
		if (DoSound)
		{
			PORTD |= (1<<PD5); //DEBUG
		}
	}

    dcf77_bit = bit;
}

void CheckDCF77Sync(void)
{
	uint8_t bit = dcf77_bit;


	// Sind 10ms vergangen, also dcf_bit in { 0, 1 }?
	if (bit < 2)
	{

		// Ja:
		// Schnappschuss zurcksetzen
		dcf77_bit = 0xff;

		// Die Zeit tickt 10ms weiter
		dcf77_tick (&dcf77, bit);

		// DCF-Abgleich erfolgreich?
		if (dcf77.time_changed)
		{
			// Neue Zeit ist in time angekommen!
			// Merker zurcksetzen
			dcf77.time_changed = 0;

			// Zeit in dcf77.ptime auswerten
			// ...
		}
	}
}